from enum import Enum
from pathlib import Path
from typing import Literal, Tuple, Union

from ...math_signals import Relation
from ...named_signals import NamedRelation
from .write_mat import write_mat_file
from .write_txt import write_txt_file


class FileType(Enum):
    MAT = ".mat"
    TXT = ".txt"


def write_data(
    relation: Union[Relation, NamedRelation],
    file_name: Union[Path, str] = None,
    header: Tuple[str, str] = None,
    file_type: Literal[".mat", ".txt"] = None,
) -> None:

    """This is a common function to write data."""

    if isinstance(file_name, str):
        file_name = Path(file_name)

    suffix = file_name.suffix

    if suffix == FileType.MAT.value or file_type == FileType.MAT.value:
        write_mat_file(relation, file_name, header)
    else:
        write_txt_file(relation, file_name, header)



import unittest

import numpy as np
from scipy.integrate import cumulative_trapezoid
from numpy.testing import assert_array_equal

from ..math_relation import Relation
from ..defaults.base_structures import BaseXY


def pre_integr(x, y):
    return BaseXY(x=x[1:], y=np.ones(x[1:].size))


class PreTestRelation:

    # @staticmethod
    # def pre_test_input(test_case: unittest.TestCase, relation_class: Relation):

    #     input_data = [([1, 2, 3, 4, 5], [10, 20, 30, 40, 50]),
    #         (np.array([1, 2, 3, 4, 5]), np.array([10, 20, 30, 40, 50])),
    #         (np.array([1, 2, 3, 4, 5]), [10, 20, 30, 40, 50]),
    #         ([1, 2, 3, 4, 5], np.array([10, 20, 30, 40, 50]))
    #                   ]
    #     for k in input_data:
    #         with test_case.subTest(k=k):
    #             r = relation_class(*k)  # type: Relation
    #             result_x, result_y = r.get_data()
    #             assert_array_equal(result_x, np.array(k[0]))
    #             assert_array_equal(result_y, np.array(k[1]))

    #     input_data_2 = [[[1, 2, 3, 4, 5], [10, 20, 30, 40, 50]],
    #                     ([1, 2, 3, 4, 5], [10, 20, 30, 40, 50]),
    #                     {'x': [1, 2, 3, 4, 5], 'y':[10, 20, 30, 40, 50]}]

    #     for k in input_data_2:
    #         with test_case.subTest(k=k):
    #             r = relation_class(k)  # type: Relation
    #             result_x, result_y = r.get_data()
    #             if isinstance(k, dict):
    #                 k = list(k.values())
    #             assert_array_equal(result_x, np.array(k[0]))
    #             assert_array_equal(result_y, np.array(k[1]))

    @staticmethod
    def pre_test_input(test_case: unittest.TestCase, relation_class: Relation):

        input_data = [
            ([1, 2, 3, 4, 5], [10, 20, 30, 40, 50]),
            (np.array([1, 2, 3, 4, 5]), np.array([10, 20, 30, 40, 50])),
            (np.array([1, 2, 3, 4, 5]), [10, 20, 30, 40, 50]),
            ([1, 2, 3, 4, 5], np.array([10, 20, 30, 40, 50])),
        ]
        for k in input_data:
            with test_case.subTest(k=k):
                r = relation_class(*k)  # type: Relation
                result_x, result_y = r.get_data()
                assert_array_equal(result_x, np.array(k[0]))
                assert_array_equal(result_y, np.array(k[1]))

    @staticmethod
    def pre_test_math(test_case: unittest.TestCase, relation_class: Relation):

        x = np.array([1, 2, 3, 4, 5], dtype="float")
        y1 = np.array([10, 20, 30, 40, 50], dtype="float")

        y2 = np.array([2, 4, 6, 8, 10], dtype="float")

        r1 = relation_class(x, y1)  # type: Relation
        r2 = relation_class(x, y2)  # type: Relation
        operation = ["__add__", "__sub__", "__mul__", "__truediv__", "__pow__"]

        for m in operation:
            for k in [(r2, y2), (2, 2)]:
                with test_case.subTest(k=k, m=m):
                    math_check(r1, k[0], x, y1, k[1], m)

    @staticmethod
    def pre_test_integrate_diff(test_case: unittest.TestCase, relation_class: Relation):

        x = np.array([1, 2, 3, 4, 5, 6])
        y = np.array([10, 20, 30, 40, 50, 60])

        r = relation_class(x, y)  # type: Relation
        dr = r.diff()
        dx, dy = dr.get_data()
        assert_array_equal(dx, x[:-1] + (r.dx) / 2)
        assert_array_equal(dy, np.diff(y))

        d2r = r.diff().diff()
        d2x, d2y = d2r.get_data()
        assert_array_equal(d2x, x[:-2] + r.dx)
        assert_array_equal(d2y, np.diff(np.diff(y)))

        ir = r.integrate()
        ix, iy = ir.get_data()
        assert_array_equal(ix, x[1:])
        assert_array_equal(iy, cumulative_trapezoid(y))

        i2r = r.integrate().integrate()
        i2x, i2y = i2r.get_data()
        assert_array_equal(x[2:], i2x)
        assert_array_equal(i2y, cumulative_trapezoid(cumulative_trapezoid(y)))

        same_r = r.integrate().integrate().diff().diff()
        same_x, same_y = same_r.get_data()
        assert_array_equal(same_x, x[3:-1])
        assert_array_equal(same_y, y[3:-1])

        same_r2 = r.integrate().diff().integrate().diff()
        same_x2, same_y2 = same_r2.get_data()
        assert_array_equal(same_x2, x[3:-1])
        assert_array_equal(same_y2, y[3:-1])

    @staticmethod
    def pre_test_interpolate_extrapolate(
        test_case: unittest.TestCase, relation_class: Relation
    ):
        x = [1, 2, 3, 4, 5]
        y = [10, 20, 30, 40, 50]
        r = relation_class(x, y)  # type: Relation
        new_x = [1.1, 2.1, 3.1, 4.1, 5.1]
        new_r = r.interpolate_extrapolate(new_x)
        new_x2, new_y = new_r.get_data()
        assert_array_equal(new_x, new_x2)
        assert_array_equal(new_y, np.array([11.0, 21.0, 31.0, 41.0, 0.0]))

    @staticmethod
    def pre_test_convolve_correlate(
        test_case: unittest.TestCase, r1, r2, relation_class: Relation
    ):

        rconv = relation_class.convolve(r1, r2)
        test_case.assertIsInstance(rconv, relation_class)

        rcorr = relation_class.correlate(r1, r2)
        test_case.assertIsInstance(rcorr, relation_class)

    # @staticmethod
    # def pre_test_names(test_case: unittest.TestCase, relation_class: Relation, default_name: str):
    #     r = relation_class([0, 1, 2], [0, 1, 2])  # type: Relation
    #     test_case.assertEqual(r.get_name(), '{0}{1}'.format(default_name, r._quantity-1))
    #     name = 'test'
    #     r2 = relation_class([0, 1, 2], [1, 2, 3], name)  # type: Relation
    #     test_case.assertEqual(r2.get_name(), name)


class TestRelation(unittest.TestCase):
    pre_cases = PreTestRelation

    def test_input(self):
        self.pre_cases.pre_test_input(self, Relation)

    def test_math(self):
        self.pre_cases.pre_test_math(self, Relation)

    def test_integrate_diff(self):
        self.pre_cases.pre_test_integrate_diff(self, Relation)

    def test_interpolate_extrapolate(self):
        self.pre_cases.pre_test_interpolate_extrapolate(self, Relation)

    def test_conv_corr(self):
        r1 = Relation([1, 2, 3, 4, 5], [10, 20, 30, 40, 50])
        r2 = Relation([1, 2, 3, 4, 5], [10, 20, 30, 40, 50])
        self.pre_cases.pre_test_convolve_correlate(self, r1, r2, Relation)


def math_check(r1, ry, x, y1, y, operation):
    r = r1.__getattribute__(operation)(ry)  # type: Relation
    if operation != "__pow__":
        y = y1.__getattribute__(operation)(y)
    else:
        if isinstance(y1, np.ndarray):
            y = np.abs(y1).__getattribute__(operation)(y) * np.sign(y1)
        else:
            y = y1.__getattribute__(operation)(y)
    result_x, result_y = r.get_data()
    assert_array_equal(result_x, x)
    assert_array_equal(result_y, y)



from ..framemakers.ipywidgetframer.figwidget.bokehfig import BokehFigureMaker

from ..framemakers.ipywidgetframer.ipywidgetframe import WidgetFrameMaker

from .common_view_builder import CommonVeiwSweepBuilder

from ..framemakers.ipywidgetframer.default_ipywidget_properties import (
    get_checkbox_frame,
    get_dropdown_frame,
    get_figure_frame,
)
from ..framemakers.ipywidgetframer.figwidget.default_bokeh_properties import (
    get_figure_window,
)


def get_view_sweep_bokeh_ipywidget(
    k_width=1.0,
    k_height=1.0,
):
    return CommonVeiwSweepBuilder(
        BokehFigureMaker,
        WidgetFrameMaker,
        get_figure_window(
            "Signal", "Time, s", "Amplitude", 0.7 * k_width, 0.4 * k_height
        ),
        get_figure_frame(0.7 * k_width, 0.4 * k_height),
        get_checkbox_frame("Signals", "No signal", 0.2 * k_width, 0.4 * k_height),
        get_figure_window(
            "Amplitude spectrum",
            "Frequency, Hz",
            "Amplitude",
            0.7 * 0.5 * k_width,
            0.4 * k_height,
        ),
        get_figure_frame(0.7 * 0.5 * k_width, 0.4 * k_height),
        get_figure_window(
            "Phase spectrum",
            "Frequency, Hz",
            "Phase",
            0.7 * 0.5 * k_width,
            0.4 * k_height,
        ),
        get_figure_frame(0.7 * 0.5 * k_width, 0.4 * k_height),
        get_figure_window(
            "Sweep", "Time, s", "Amplitude", 0.7 * k_width, 0.4 * k_height
        ),
        get_figure_frame(0.7 * k_width, 0.4 * k_height),
        get_figure_window(
            "Spectrogram", "Time, s", "Frequency, Hz", 0.7 * k_width, 0.4 * k_height
        ),
        get_figure_frame(0.7 * k_width, 0.4 * k_height),
        get_dropdown_frame("Sweeps", "No sweep"),
        get_figure_window(
            "Envelop", "Time, s", "Amplitude", 0.7 * k_width, 0.4 * k_height
        ),
        get_figure_frame(0.7 * k_width, 0.4 * k_height),
        get_checkbox_frame("Sweeps", "No sweep", 0.2 * k_width, 0.4 * k_height),
    )



from typing import Type, Union

from ...named_signals.named_signal import NamedSignal
from ...named_signals.named_sweep import NamedSweep

from ..base_view.abc_common_framer_grapher.figure import CommonFigure, CommonFigureMaker
from ..base_view.abc_common_framer_grapher.frame import CommonFrameMaker

from .controller import InteracterPlotViewSweep


class NotCorrectDataError(Exception):
    pass


class CommonVeiwSweepBuilder:
    def __init__(
        self,
        grapher: Type[CommonFigureMaker],
        framer: Type[CommonFrameMaker[CommonFigure, InteracterPlotViewSweep]],
        properties_signal_figure,
        properties_signal_frame,
        properties_checkbox_signals_frame,
        properties_freq_amplitude_figure,
        properties_freq_amplitude_frame,
        properties_freq_phase_figure,
        properties_freq_phase_frame,
        properties_sweep_figure,
        properties_sweep_frame,
        properties_spectrogram_figure,
        properties_spectrogram_frame,
        properties_dropdown_images_spectrogram_frame,
        properties_envelop_figure,
        properties_envelop_frame,
        properties_checkbox_sweep_frame,
    ) -> None:

        self._grapher = grapher()
        self._controller = InteracterPlotViewSweep()

        signal_figure = grapher.get_figure(properties_signal_figure)
        self._signal_frame = framer.get_figure_frame(
            signal_figure, properties_signal_frame
        )

        self._checkbox_signals_frame = framer.get_checkbox_frame(
            properties_checkbox_signals_frame, self._controller
        )

        spectrum_amplitude_figure = grapher.get_figure(properties_freq_amplitude_figure)
        self._spectrum_amplitude_frame = framer.get_figure_frame(
            spectrum_amplitude_figure, properties_freq_amplitude_frame
        )

        spectrum_phase_figure = grapher.get_figure(properties_freq_phase_figure)
        self._spectrum_phase_frame = framer.get_figure_frame(
            spectrum_phase_figure, properties_freq_phase_frame
        )

        sweep_figure = grapher.get_figure(properties_sweep_figure)
        self._sweep_frame = framer.get_figure_frame(
            sweep_figure, properties_sweep_frame
        )

        spectrogram_images_figure = grapher.get_figure(properties_spectrogram_figure)
        self._spectrogram_images_frame = framer.get_figure_frame(
            spectrogram_images_figure, properties_spectrogram_frame
        )

        self._dropdown_images_spectrogram = framer.get_dropdown_frame(
            properties_dropdown_images_spectrogram_frame, self._controller
        )

        envelop_images_figure = grapher.get_figure(properties_envelop_figure)
        self._envelop_frame = framer.get_figure_frame(
            envelop_images_figure, properties_envelop_frame
        )

        self._checkbox_sweep_frame = framer.get_checkbox_frame(
            properties_checkbox_sweep_frame, self._controller
        )

        spectrum_section = self._spectrum_amplitude_frame.h_add(
            self._spectrum_phase_frame
        )
        result_frame = self._signal_frame.v_add(spectrum_section)
        result_frame = result_frame.h_add(self._checkbox_signals_frame)
        sweep_section = (
            self._sweep_frame.v_add(self._dropdown_images_spectrogram)
            .v_add(self._spectrogram_images_frame)
            .v_add(self._envelop_frame)
            .h_add(self._checkbox_sweep_frame)
        )

        result_frame = result_frame.v_add(sweep_section)

        self._result_frame = result_frame

        self._controller.set_controlled_object(
            self._signal_frame,
            self._checkbox_signals_frame,
            self._spectrum_amplitude_frame,
            self._spectrum_phase_frame,
            self._sweep_frame,
            self._dropdown_images_spectrogram,
            self._spectrogram_images_frame,
            self._envelop_frame,
            self._checkbox_sweep_frame,
        )

    def add(self, data: Union[NamedSignal, NamedSweep]):

        line_properties = self._grapher.get_next_line_properties()
        checkbox_properties = {"color": line_properties["line_color"]}
        image_properties = self._grapher.get_next_image_properties()

        name = str(data)

        if isinstance(data, NamedSweep):
            self._sweep_frame.get_figure().add_line(
                *data.get_data(), name, None, line_properties
            )

            spectrogram_figure = self._spectrogram_images_frame.get_figure()

            s = data.spectrogram
            spectrogram_figure.add_image(
                s.time, s.frequency, s.spectrogram_matrix, name, None, image_properties
            )
            spectrogram_figure.add_line(
                *data.f_t.get_data(), name, None, line_properties
            )
            self._dropdown_images_spectrogram.add_choise(name)

            self._envelop_frame.get_figure().add_line(
                *data.a_t.get_data(), name, None, line_properties
            )
            self._checkbox_sweep_frame.add_choise(
                name, None, checkbox_properties, self._controller
            )

        elif isinstance(data, NamedSignal):

            self._signal_frame.get_figure().add_line(
                *data.get_data(), name, None, line_properties
            )
            self._checkbox_signals_frame.add_choise(
                name, None, checkbox_properties, self._controller
            )

        else:
            raise NotCorrectDataError

        amplitude_spectrum = data.get_amplitude_spectrum(name=name)
        self._spectrum_amplitude_frame.get_figure().add_line(
            *amplitude_spectrum.get_data(),
            str(amplitude_spectrum),
            None,
            line_properties
        )
        phase_spectrum = data.get_phase_spectrum(name=name)
        self._spectrum_phase_frame.get_figure().add_line(
            *phase_spectrum.get_data(), str(phase_spectrum), None, line_properties
        )

    def show(self):
        return self._result_frame.get_output()



import numpy as np

from ..math_relation import Relation
from ..math_sweep import Sweep
from ..math_uncalcsweep import ApriorUncalculatedSweep
from ..utility_functions.ftat_functions import dwell
from ..utility_functions.tukey import tukey_a_t


def get_dwell_sweep(
    t: np.ndarray,
    f_start=1.0,
    f_central=5.0,
    f_end=100.0,
    t_tapper=1.0,
    aprior_data: Relation = None,
) -> Sweep:
    """Create a dwell sweep using an a priori data.

    t_tapper in seconds is used to apply tukey function at the end of dwell sweep signal.
    """
    ftat_method = dwell(f_start, f_end, f_central)
    tukey_coef = np.append(
        np.ones(int(len(t) / 2)), tukey_a_t(t, t_tapper)[int(len(t) / 2) + 1 :]
    )
    uasw = ApriorUncalculatedSweep(t, aprior_data, ftat_method)

    return tukey_coef * uasw()



from pathlib import Path
from typing import Union
from ...config import Config
from ...named_signals import NamedRelation


class ExtensionFileError(Exception):
    pass


def get_path_file_name(relation, file_name: Union[Path, str] = None, suffix="") -> Path:

    if file_name is None:
        if isinstance(relation, NamedRelation):
            file_name = f"{str(relation)}{suffix}"
        else:
            file_name = f"unnamed{suffix}"

    if isinstance(file_name, str):
        file_name = Path(file_name)

    if not file_name.is_absolute():
        file_name = Config.DEFAULT_PATH / file_name

    if file_name.suffix == "":
        file_name = file_name.parent / f"{file_name.name}{suffix}"

    if file_name.suffix != suffix:
        raise ExtensionFileError(
            f"The subbmited file_name {file_name} do not match extansion {suffix}"
        )

    return file_name



"""This is where defualt methods are defined."""

from typing import Callable, Union, Any, TYPE_CHECKING, Type, Tuple

import numpy as np
from scipy.interpolate import interp1d  # type: ignore
from scipy.integrate import cumulative_trapezoid, trapz  # type: ignore

from .base_structures import NotEqualError, MathOperation

x = np.ndarray
y = np.ndarray
frequency = np.ndarray
spectrum = np.ndarray
time = np.ndarray
amplitude = np.ndarray

if TYPE_CHECKING:
    from math_signals.math_relation import Relation

Number = Union[float, int, complex]


def math_operation(
    x: np.ndarray,
    y1: np.ndarray,
    y2: Union[np.ndarray, Number],
    name_operation: MathOperation,
) -> Tuple[x, y]:
    """Math operations.

    Using numpy math operations.
    """
    if name_operation == MathOperation.POW:
        y = np.abs(y1).__getattribute__(name_operation.value)(y2) * np.sign(y1)
    else:
        y = y1.__getattribute__(name_operation.value)(y2)

    return x, y


def one_integrate(y: np.ndarray, x: np.ndarray = None) -> float:
    """Integration.

    Taking the integral on a segment. Return of the area under the graph.
    """
    return trapz(y, x)


def integrate(x: np.ndarray, y: np.ndarray, dx: float = None) -> Tuple[x, y]:
    """Integration.

    Using the scipy.integrate.cumtrapz function.
    """
    if dx is None:
        dx = x[1] - x[0]
    return x[1:], cumulative_trapezoid(y) * (dx)


def differentiate(x: np.ndarray, y: np.ndarray, dx: float = None) -> Tuple[x, y]:
    """Differentiation.

    Using the numpy.diff function."""
    if dx is None:
        dx = x[1] - x[0]
    return x[:-1] + (dx) / 2, np.diff(y) / (dx)


def interpolate_extrapolate(
    x: np.ndarray, y: np.ndarray, bounds_error=False, fill_value=0.0
) -> Callable[[x], y]:
    """Interpolation.

    Using the scyipy.interpolate.interp1d function.
    Returning function of interpolation.
    """
    return interp1d(x, y, bounds_error=bounds_error, fill_value=fill_value)


def get_common_x(
    x1: np.ndarray, x2: np.ndarray, dx1: float = None, dx2: float = None
) -> x:
    """Specifies the overall x-axis.

    Finds the general sample rate and beginning and end of sequence.
    """
    if dx1 is None:
        dx1 = x1[1] - x1[0]
    if dx2 is None:
        dx2 = x2[1] - x2[0]
    dx = dx1 if dx1 <= dx2 else dx2
    x_start = x1[0] if x1[0] <= x2[0] else x2[0]
    x_end = x1[-1] if x1[-1] >= x2[-1] else x2[-1]
    X = x_end - x_start
    return np.linspace(x_start, (int(X / dx)) * dx, int(X / dx) + 1)


def correlate(cls: Type["Relation"], r1: "Relation", r2: "Relation") -> Tuple[x, y]:
    """Correlation.

    Using the numpy.correlate function.
    """
    r1 = r1.shift(-r1._x[0])
    r2 = r2.shift(-r2._x[0])
    r1, r2 = cls.equalize(r1, r2)
    x, y1 = r1.get_data()
    _, y2 = r2.get_data()
    return np.append(np.sort(-1 * x)[:-1], x), np.correlate(y1, y2, "full")


def convolve(cls: Type["Relation"], r1: "Relation", r2: "Relation") -> Tuple[x, y]:
    """Convolution.

    Using the numpy.convlove function.
    """
    r1 = r1.shift(-r1._x[0])
    r2 = r2.shift(-r2._x[0])
    r1, r2 = cls.equalize(r1, r2)
    x, y1 = r1.get_data()
    _, y2 = r2.get_data()
    return np.append(np.sort(-1 * x)[:-1], x), np.convolve(y1, y2, "full")


# ==============================================================================


def signal2spectrum(
    time: np.ndarray, amplitude: np.ndarray, is_start_zero=False
) -> Tuple[frequency, spectrum]:
    """Forward Fourier Transform.

    Using the numpy.fft.rfft function.
    """
    if not is_start_zero:
        if time[0] > 0.0:
            time = np.linspace(0.0, time[-1], int(time[-1] / (time[1] - time[0])) + 1)
            amplitude = np.append(np.zeros(time.size - amplitude.size), amplitude)
        elif time[-1] < 0.0:
            time = np.linspace(
                time[0], 0.0, int(abs(time[0]) / (time[1] - time[0])) + 1
            )
            amplitude = np.append(amplitude, np.zeros(time.size - amplitude.size))

        amplitude = np.append(amplitude[time >= 0.0], amplitude[time < 0.0])

    spectrum = np.fft.rfft(amplitude)
    frequency = np.fft.rfftfreq(
        amplitude.size, d=(time[-1] - time[0]) / (amplitude.size)
    )
    return frequency, spectrum


def spectrum2sigmal(
    frequency: np.ndarray, spectrum: np.ndarray, time_start: float = None
) -> Tuple[time, amplitude]:
    """Inverse Fourier Transform.

    Using the numpy.fft.irfft function.
    """
    amplitude = np.fft.irfft(spectrum)  # type: np.ndarray
    if time_start is None:
        time = np.linspace(
            0,
            (amplitude.size - 1) / (2 * (frequency[-1] - frequency[0])),
            amplitude.size,
        )
    else:
        time = np.linspace(
            time_start,
            time_start + (amplitude.size - 1) / (2 * (frequency[-1] - frequency[0])),
            amplitude.size,
        )
        amplitude = np.append(amplitude[time >= 0.0], amplitude[time < 0.0])

    return time, amplitude



from .view import GeneralView
from ..framemakers.ipywidgetframer.figwidget.bokehfig import BokehFigureMaker
from ..framemakers.ipywidgetframer.figwidget.matplotlibfig import MatplotlibFigureMaker
from ..framemakers.ipywidgetframer.ipywidgetframe import WidgetFrameMaker

from ..framemakers.ipywidgetframer.default_ipywidget_properties import (
    get_checkbox_frame,
    get_dropdown_frame,
    get_figure_frame,
)
from ..framemakers.ipywidgetframer.figwidget.default_bokeh_properties import (
    get_figure_window as get_bokeh_figure_window,
)
from ..framemakers.ipywidgetframer.figwidget.default_matplotlib_properties import (
    get_figure_window as get_matplotlib_figure_window,
)


def get_general_view_bokeh_ipywidget(
    x_axis_label: str, y_axis_label: str, title: str, k_width: float, k_height: float
):
    return GeneralView(
        BokehFigureMaker,
        WidgetFrameMaker,
        get_bokeh_figure_window(
            title, x_axis_label, y_axis_label, 0.8 * k_width, k_height * 0.9
        ),
        get_figure_frame(k_width=0.8 * k_width, k_height=k_height * 0.9),
        get_checkbox_frame(title, "No data", 0.2 * k_width, k_height * 0.9),
        get_dropdown_frame("Data:", "No data!"),
    )


def get_general_view_matplotlib_ipywidget(
    x_axis_label: str,
    y_axis_label: str,
    title: str,
    k_width: float = 1.0,
    k_height: float = 1.0,
):

    return GeneralView(
        MatplotlibFigureMaker,
        WidgetFrameMaker,
        get_matplotlib_figure_window(
            title, x_axis_label, y_axis_label, 0.8 * k_width, k_height * 0.9
        ),
        get_figure_frame(k_width=0.8 * k_width, k_height=k_height * 0.9),
        get_checkbox_frame(title, "No data", 0.2 * k_width, k_height * 0.9),
        get_dropdown_frame("Data:", "No data!"),
    )



from typing import Any, Callable
from abc import ABCMeta, abstractmethod


class CommonController(metaclass=ABCMeta):
    @abstractmethod
    def widget_notify(self, sender: Any) -> Callable[[Any], None]:
        pass



"""The module contains simple functions for loading and saving data.

Works with next formats:
> * .txt - Methods: `sweep_design.dataio.read.read_txt.read_txt_file`, 
    `sweep_design.dataio.write.write_txt.write_txt_file`, 
    `sweep_design.dataio.write.write_data.write_data`
> * .mat - Methods `sweep_design.dataio.read.read_mat.read_mat_file`, 
    `sweep_design.dataio.write.write_mat.write_mat_file`, 
    `sweep_design.dataio.write.write_data.write_data`

For `sweep_design.dataio.write.write_data.write_data`, if the format is 
different or exсluded, then it will be enterpreted as a txt file. 

For `file_name` you can use both absolute path and relative path.
If you use a raltive path, then the absolut path will consist from two parts:
> first part is the default path **DEFAULT_PATH** from the config file from class 
`sweep_design.config.config.Config`,  
> second part is the relative path.

That is working to write and read.

You can change **DEFAULT_PATH** in config file.

If `file_name` is not specified for the wroten file, then the default name 
`unnamed.(extension you chose)` will be used for 
`sweep_design.math_signals.math_relation.Relation` instance you want save. 
If this is instance of  
`sweep_design.named_signals.named_relation.NamedRelation` you want to store, 
then the name will be `str(instance of NamedRelation).(extension you chose)`. 

"""

from .read.read_mat import read_mat_file as read_mat_file
from .read.read_txt import read_txt_file as read_txt_file

from .write.write_data import write_data as write_data
from .write.write_mat import write_mat_file as write_mat_file
from .write.write_txt import write_txt_file as write_txt_file



import numpy as np

from ..math_uncalcsweep import UncalculatedSweep
from ..math_sweep import Sweep
from ..utility_functions.tukey import tukey_a_t
from ..utility_functions import f_t_linear_array


def get_linear_sweep(t: np.ndarray, f_start=1.0, f_end=100.0, t_tapper=1.0) -> Sweep:
    """Create linear sweep.

    t_tapper in seconds is used to apply tukey function to a sweep signal.
    """

    f_t = f_t_linear_array(t, f_start, f_end)
    a_t = tukey_a_t(t, t_tapper)

    unsw = UncalculatedSweep(t, f_t, a_t)
    return unsw()



from typing import Any, Dict

from ..base_view.abc_common_framer_grapher.figure import CommonFigure

from ..base_view.abc_common_framer_grapher.frame import (
    CommonCheckBoxFrame,
    CommonDropdownFrame,
    CommonFigureFrame,
    CommonCheckBox,
)

from ..base_view.abc_common_framer_grapher.controller import CommonController


class BoundElement:
    def __init__(self, element: Any) -> None:
        self._element = element
        self._bound_elements: Dict[int, Any] = {}

    def bound_element(self, bound_element: Any) -> None:
        self._bound_elements.update({int(bound_element): bound_element})


class InteracterSource(CommonController):
    def __init__(
        self,
        dropdown_singals_frame: CommonDropdownFrame = None,
        signal_figure_frame: CommonFigureFrame[CommonFigure] = None,
        checkbox_types_signals_frame: CommonCheckBoxFrame[CommonCheckBox] = None,
    ) -> None:
        self._dropdown_singals_frame = dropdown_singals_frame
        self._signal_figure_frame = signal_figure_frame
        self._checkbox_types_signals_frame = checkbox_types_signals_frame

    def set_controlled_objects(
        self,
        dropdown_singals_frame: CommonDropdownFrame,
        signal_figure_frame: CommonFigureFrame[CommonFigure],
        checkbox_types_signals_frame: CommonCheckBoxFrame[CommonCheckBox],
    ) -> None:
        self._dropdown_singals_frame = dropdown_singals_frame
        self._signal_figure_frame = signal_figure_frame
        self._checkbox_types_signals_frame = checkbox_types_signals_frame

    def widget_notify(self, obj):
        def listner(value: Dict[str, Any]):
            if obj is self._dropdown_singals_frame:
                self._call_dropdown_signal(value)

            if obj in self._checkbox_types_signals_frame.get_checkboxs().values():
                self._call_checkbox_types_signals(obj.get_name(), value)

                # line = self._signal_figure_frame.get_figure().get_line(obj.get_name())
                # print(line)
                # print(obj.get_name())
                # if line:
                #     line.set_visible(value['new'])

        return listner

    def _call_checkbox_types_signals(self, group: str, value: bool):
        current_dropdown_choise = self._dropdown_singals_frame.get_choise()

        for line in self._signal_figure_frame.get_figure().get_lines_by_groups(
            set([group, current_dropdown_choise])
        ):
            line.set_visible(value["new"])
        if group == "limits":
            for line in self._signal_figure_frame.get_figure().get_lines_by_groups(
                set([group])
            ):
                line.set_visible(value["new"])

    def _call_dropdown_signal(self, value: Dict[str, Any]):
        value_new = value["new"]
        for line in self._signal_figure_frame.get_figure().get_lines_by_groups(
            set([value_new])
        ):
            line.set_visible(True)
        value_old = value["old"]
        for line in self._signal_figure_frame.get_figure().get_lines_by_groups(
            set([value_old])
        ):
            line.set_visible(False)

        for checkbox in self._checkbox_types_signals_frame.get_checkboxs().values():
            checkbox.set_select(True)



from typing import Any, Protocol, Sized, Tuple, TypeVar, Union, Type

import numpy as np

Num = Union[float, int, complex]
R = TypeVar("R", bound="MathRelation")
R2 = TypeVar("R2", bound="MathRelation")
SP = TypeVar("SP", bound="MathSpectrum")
S = TypeVar("S", bound="MathSignal")
SPRN = Union["MathSpectrum", "MathRelation", Num]
SSPR = Union["MathSpectrum", "MathSignal", "MathRelation"]
SSPRN = Union["MathSpectrum", "MathSignal", "MathRelation", Num]


class MathRelation(Sized, Protocol):
    _x: np.ndarray
    _y: np.ndarray

    def get_data(self) -> Tuple[np.ndarray, np.ndarray]:
        ...

    def max(self) -> Num:
        ...

    def min(self) -> Num:
        ...

    def get_norm(self) -> float:
        ...

    def select_data(self: R, x_start: Num = None, x_end: Num = None, **kwargs) -> R:
        ...

    def exp(self: R, **kwargs) -> R:
        ...

    def diff(self: R, **kwargs) -> R:
        ...

    def integrate(self: R, **kwargs) -> R:
        ...

    def interpolate_extrapolate(self: R, new_x: np.ndarray, **kwargs) -> R:
        ...

    def shift(self: R, x_shift: Num = 0, **kwargs) -> R:
        ...

    def __add__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __radd__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __sub__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __rsub__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __mul__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __rmul__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __truediv__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __rtruediv__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __pow__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __rpow__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __iadd__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __isub__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __imul__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __idiv__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    def __ipow__(self: R, other: Union["MathRelation", Num], **kwargs) -> R:
        ...

    @staticmethod
    def equalize(r1: R, r2: R2) -> Tuple[R, R2]:
        ...

    @classmethod
    def correlate(cls: Type[R], r1: "MathRelation", r2: "MathRelation", **kwargs) -> R:
        ...

    @classmethod
    def convolve(cls: Type[R], r1: "MathRelation", r2: "MathRelation", **kwargs) -> R:
        ...


class MathSpectrum(MathRelation, Protocol):

    spectrum: Any

    def get_signal(self, recalculate=False, start_time: float = None) -> "MathSignal":
        ...

    def get_amp_spectrum(self: R, **kwargs) -> "MathRelation":
        ...

    def get_phase_spectrum(self: R, **kwargs) -> "MathRelation":
        ...

    def get_reverse_filter(
        self: SP,
        percent: Union[float, int] = 5.0,
        subtrack_phase=True,
        f_start: float = None,
        f_end: float = None,
        **kwargs
    ) -> SP:
        ...

    def add_phase(self: SP, other: SSPR, **kwargs) -> SP:
        ...

    def sub_phase(self: SP, other: SSPR, **kwargs) -> SP:
        ...

    @classmethod
    def get_spectrum_from_amp_phase(
        cls: Type[SP], s1: MathRelation, s2: MathRelation, **kwargs
    ) -> SP:
        ...

    @classmethod
    def convolve(cls: Type[SP], r1: SSPR, r2: SSPR, **kwargs) -> SP:
        ...

    @classmethod
    def correlate(cls: Type[SP], r1: SSPR, r2: SSPR, **kwargs) -> SP:
        ...

    def __add__(self: SP, a: SSPRN, **kwargs) -> SP:
        ...

    def __sub__(self: SP, a: SSPRN, **kwargs) -> SP:
        ...

    def __mul__(self: SP, a: SSPRN, **kwargs) -> SP:
        ...

    def __truediv__(self: SP, a: SSPRN, **kwargs) -> SP:
        ...

    def __pow__(self: SP, a: SSPRN, **kwargs) -> SP:
        ...


class MathSignal(MathRelation, Protocol):
    def get_spectrum(self, recalculate=False, is_start_zero=False) -> "MathSpectrum":
        ...

    def get_reverse_signal(
        self: S,
        percent: Union[float, int] = 5.0,
        subtrack_phase: bool = True,
        f_start: float = None,
        f_end: float = None,
        **kwargs
    ) -> S:
        ...

    def add_phase(self: S, other: SSPR, **kwargs) -> S:
        ...

    def sub_phase(self: S, other: SSPR, **kwargs) -> S:
        ...

    @classmethod
    def convolve(cls: Type[S], r1: SSPR, r2: SSPR, **kwargs) -> S:
        ...

    @classmethod
    def correlate(cls: Type[S], r1: SSPR, r2: SSPR, **kwargs) -> S:
        ...

    def __add__(self: S, a: SSPRN, **kwargs) -> S:
        ...

    def __sub__(self: S, a: SSPRN, **kwargs) -> S:
        ...

    def __mul__(self: S, a: SSPRN, **kwargs) -> S:
        ...

    def __truediv__(self: S, a: SSPRN, **kwargs) -> S:
        ...

    def __pow__(self: S, a: SSPRN, **kwargs) -> S:
        ...


class MathSpectrogram(Protocol):
    t: np.ndarray
    f: np.ndarray
    S: np.ndarray


class MathSweep(MathSignal, Protocol):
    f_t: MathRelation
    a_t: MathRelation
    spectrogram: MathSpectrogram
    aprior_signal: MathSignal



from ..math_signals.defaults import sweep_methods as dfsm


class SweepConfig:

    """The configuratuin file for sweep.

    > Several methods are defined for default calculations:

    **spectrogram_method_default**
    > The method by which the spectrogram will be calculated.
    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.sweep_methods.get_spectrogram`

    > **input**:
    >> **time**: `numpy.ndarray`
    **amplitude**: np.ndarry
    **dt**: float

    > **output**:
    >> `Tuple` [
    >> **time**: `numpy.ndarray`,
    >> **frequency**: `numpy.ndarray`,
    >> **spectrogram**: `numpy.ndarray`
    ]

    - - -

    **get_f_t**
    > The method by which frequency versus time will be calculated.
    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.sweep_methods.get_f_t`

    > **input**:
    >> **time**: `numpy.ndarray`
    **amplitude**: `numpy.ndarray`

    > **output**:
    >> `sweep_design.math_signals.math_relation.Relation`

    - - -

    **get_a_t**
    > The method by which the time envelope of the signal will be calculated.
    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.sweep_methods.get_a_t`

    > **input**:
    >> **time**: `numpy.ndarray`
    **amplitude**: `numpy.ndarray`

    > **output**:
    >> `sweep_design.math_signals.math_relation.Relation`

    - - -

    **integrate_function**
    > A method for integrating a function rather than a sequence.
    Method derived from default function:
    >> `sweep_design.math_signals.defaults.sweep_methods.integrate_quad`

    > **input**:
    >> **f_t_function**: `Callable`[[**time**: `numpy.ndarray`],
    **frequency**: `numpy.ndarray`]
    **time**: `numpy.ndarray`

    > **output**:
    >> **y**: `numpy.ndarray`

    - - -

    **freq2time**
    > The simple method to extract the time envelope of a sweep signal and
    the time-frequency function to generate a sweep signal from a priori data.
    Method derived from default function:
    >> `sweep_design.math_signals.defaults.sweep_methods.simple_freq2time`

    > **input**:
    >> **spectrum**: `sweep_design.math_signals.math_signal.Spectrum`

    > **output**:
    >> Tuple [
    >> **time**: `numpy.ndarray`,
    **frequency**: `numpy.ndarray`,
    **envelope**: `numpy.ndarray`
    ]

    - - -

    The above methods can be overridden with your own here, or you can import the
    class SweepConfig somewhere and override it there.
     (They must be written according to the rules corresponding to
     the input and output parameters)
    """

    # Methods for Sweep.
    spectrogram_method = dfsm.get_spectrogram
    get_f_t = dfsm.get_f_t
    get_a_t = dfsm.get_a_t

    # Methods for UncalculatedSweep.
    integrate_function = dfsm.integrate_quad

    # Method for ApriorUncalculatedSweep.
    freq2time = dfsm.simple_freq2time



from typing import Union, Callable, Optional

from ..base_header import HeaderBase

Num = Union[float, int, complex]
InName = Union[HeaderBase, str, Callable[[], str]]


def make_name(name: Optional[InName], func: Callable[..., str], *args) -> InName:

    if name is not None:
        return name

    def call() -> str:
        return func(*args)

    return call


def make_category(
    named_relation: HeaderBase,
    category: Optional[str],
    getter_func: Callable[[str], str] = None,
) -> str:

    if category is None:
        if getter_func is None:
            return named_relation.category
        else:
            return getter_func(named_relation.category)
    return category



from pathlib import Path
from ..math_signals.defaults import methods as dfm


class Config:
    """The configuratuin file.

    The methods used to calculate the various parameters are provided in the
    configuration file. Also the file contain an other parameters.
    -----------------------------------------------------------------------

    **DEFAULT_PATH** - This is the path using to write and read data.
    This directory defaults to the current working derectiory, from where
    functions are called.

    - - -

    **CONVERT2ARRAY** - If parameter is True, then converted **input** array to `numpy.ndarray`
    in the constructor of class `Relation`.

    - - -

    **interpolate_extrapolate_method**:
    > The method by which interpolation and extrapolation are performed.
    > The methods returns a function that takes a new x sequence and
    return a new y sequence.

    > **new_x** = `numpy.ndarray`
    > **new_y** = `numpy.ndarray`

    > Method derived from default function:
    >> sweep_design.math_signals.defaults.methods.interpolate_extrapolate

    > **input**:
    >> **x**: `numpy.ndarray`
    >> **y**: `numpy.ndarray`

    > **output**:
    >> `Callable`[[**new_x**], **new_y**]

    - - -

    **math_operation_method**:
    > The method of basic mathematical operations of addition (+),
    subtraction(-), multiplication (*), division(/), exponentiation (**)
    and their unary operations (+=, -=, *=, /=).

    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.methods.math_operation`

    > **input**:
    >> **x**: `numpy.ndarray`
    >> **y1**: `numpy.ndarray`
    >> **y2**: `Union`[`numpy.ndarray`, `float`, `int`, `complex`]
    >> **name_operation**: `sweep_design.math_signals.defaults.base_structures.MathOperation`

    > **output**:
    >> `Tuple` [
    >> **x**: `numpy.ndarray`,
    >> **y**: `numpy.ndarray`
    >>]

    - - -

    **integrate_one_method**:
    > Method for calculating the integral of a sequence on a segment.
    > Method derived from default function:

    >> `sweep_design.math_signals.defaults.methods.one_integrate`

    > **input**:
    >> **x**: `numpy.ndarray`
    >> **y**: `numpy.ndarray`

    > **output**:
    >> `float`

    - - -

    **integrate_method**:
    > The method by which the integration is performed.
    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.methods.integrate`

    > **input**:
    >> **x**: `numpy.ndarray`
    >> **y**: `numpy.ndarray`
    >> d**x**: float - sequence sample rate x

    > **output**:
    >> Tuple [
        **x**: `numpy.ndarray`,
        **y**: `numpy.ndarray`
    ]

    - - -

    **differentiate_method**:
    > The method by which differentiation is performed.
    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.methods.differentiate`

    > **input**:

    >> **x**: `numpy.ndarray`
    >> **y**: `numpy.ndarray`
    >> **dx**: `float` - sequence sample rate **x**

    > **output**:
    >>Tuple [
        **x**: `numpy.ndarray`,
        **y**: `numpy.ndarray`
    ]

    - - -

    **correlate_method**:
    > The method by which the correlation is performed.
    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.methods.correlate`

    > **input**:
    >> cls: Relation
    >> r1: Relation
    >> r2: Relation

    > **output**:
    >> Tuple [
        **x**: `numpy.ndarray`,
        **y**: `numpy.ndarray`
    ]

    - - -

    **convolve_method**:
    > The method by which the convolution is performed.
    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.methods.convolve`

    > **input**:
    >> cls: Relation
    >> r1: Relation
    >> r2: Relation

    > **output**:
    >> Tuple [
        **x**: `numpy.ndarray`,
        **y**: `numpy.ndarray`
    ]

    - - -

    **get_common_x**:
    > A method by which to find the common sequence of numbers along
    > the x-axis, obtained from two other sequences along the x-axis.
    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.methods.get_common_x`

    > **input**:
    >> **x1**: `numpy.ndarray`
    >> **x2**: `numpy.ndarray`
    >> **dx1**: float  - sample rate first sequence
    >> **dx2**: float  - sample rate second sequence

    > **output**:
    >> **x**: `numpy.ndarray`

    - - -

    **spectrum2signal_method**:
    > Method for converting a spectrum into a signal. (Using Fourier transform)
    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.methods.spectrum2sigmal`

    > **input**:
    >> **frequenc**: `numpy.ndarray`
    **specturm**: `numpy.ndarray`
    **start_time**: `float` = `None`

    > **output**:
    >> Tuple [
        **time**: `numpy.ndarray`,
        **amplitude**: `numpy.ndarray`
    ]

    - - -

    **signal2spectrum_method**:
    > Method for converting a signal into a spectrum. (Using Fourier transform)
    > Method derived from default function:
    >> `sweep_design.math_signals.defaults.methods.spectrum2sigmal`

    > **input**:

    >> **time**: `numpy.ndarray`
    >> **amplitude**: `numpy.ndarray`
    >> **is_start_zero** = `False`

    > **output**:
    >> Tuple [
        **frequency**: `numpy.ndarray`,
        **spectrum**: `numpy.ndarray`
    ]

    - - -

    The above methods can be overridden with your own here, or you can import the
    class `Config` somewhere and override it there.
    (They must be written according to the rules corresponding to
    the **input** and **output** parameters)

    """

    # Path for io_data module
    DEFAULT_PATH = Path(".")

    # Conver array for Relation.
    CONVERT2ARRAY = True

    # Methods for the Relation.
    interpolate_extrapolate_method = dfm.interpolate_extrapolate
    math_operation = dfm.math_operation
    integrate_one_method = dfm.one_integrate
    integrate_method = dfm.integrate
    differentiate_method = dfm.differentiate
    correlate_method = dfm.correlate
    convolve_method = dfm.convolve
    get_common_x = dfm.get_common_x

    # Methods for Spectrum and Signal.
    spectrum2signal_method = dfm.spectrum2sigmal
    signal2spectrum_method = dfm.signal2spectrum



from typing import TypeVar

import numpy as np

from ..math_relation import Relation
from ..utility_functions.emd_analyse import get_IMFs_emd
from ..utility_functions.tukey import tukey_a_t

Signal = TypeVar("Signal", bound=Relation)


def correct_sweep(signal: Signal, start_window: float = None) -> Signal:
    """Sweep correction.

    Using the EMD to subtract the last IMF from the displacement and apply
    a window in the star so that the displacement starts at zero.
    """

    displacement = signal.integrate().integrate()
    x = displacement.x

    window = tukey_a_t(x, start_window, "left")

    IMFs = get_IMFs_emd(displacement)
    new_displacement = IMFs[0] * Relation(x, window)

    signal = new_displacement.diff().diff()

    return signal


def correct_sweep_without_window(signal: Signal) -> Signal:
    """Using the EMD to subtract the last IMF from the displacement."""
    displacement = signal.integrate().integrate()
    IMFs = get_IMFs_emd(displacement)
    signal = IMFs[0].diff().diff()
    return signal



from typing import Any, List

from PyEMD import CEEMDAN, EMD
from ..math_signal import Signal


def get_IMFs_ceemdan(
    data: Signal,
    number_seedman=30,
    epsilon=0.005,
    ext_EMD: Any = None,
    parallel=False,
    processes=None,
    noise_scale=1,
    noise_kind="normal",
    range_thr=0.01,
    total_power_thr=0.05,
) -> List[Signal]:
    """Empirical mode decomposition (EMD).

    Using CEEMDAN from PyEMD (https://pyemd.readthedocs.io/) to calulate IMFs
    and return them as a list of Signals.

    About empirical mode decomposition on
    https://en.wikipedia.org/wiki/Hilbert%E2%80%93Huang_transform#Techniques
    """
    x, y = data.get_data()
    emd = CEEMDAN(
        number_seedman,
        epsilon,
        ext_EMD,
        parallel,
        processes=processes,
        noise_scale=noise_scale,
        noise_kind=noise_kind,
        range_thr=range_thr,
        total_power_thr=total_power_thr,
    )
    IMFs = emd(y)
    result = [Signal(x, k) for k in IMFs]
    return result


def get_IMFs_emd(data: Signal) -> List[Signal]:

    x, y = data.get_data()
    emd = EMD()
    IMFs = emd(y)
    result = [Signal(x, k) for k in IMFs]
    return result



from typing import Any, Dict

import numpy as np

from ..base_view.abc_common_framer_grapher.frame import CommonFrameMaker
from ..base_view.abc_common_framer_grapher.figure import CommonFigureMaker, CommonFigure

from ..view_pilot_rm_bp.controller import InteracterPilotRmBp


class PilotRmBpViewBuilder:
    def __init__(
        self,
        grapher: CommonFigureMaker,
        framer: CommonFrameMaker[CommonFigure, InteracterPilotRmBp],
        properties_freq_amp_frame: Dict[str, Any],
        properties_freq_amp_figure: Dict[str, Any],
        properties_freq_phase_frame: Dict[str, Any],
        properties_freq_phase_figure: Dict[str, Any],
        properties_checkbox_signals_frame: Dict[str, Any],
        properties_checkbox_type_signals_frame: Dict[str, Any],
        properties_main_sweep_frame: Dict[str, Any],
        properties_main_sweep_figure: Dict[str, Any],
        properties_dropdown_images_spectrogram_frame: Dict[str, Any],
        properties_spectrogram_frame: Dict[str, Any],
        properties_spectrogram_figure: Dict[str, Any],
    ) -> None:

        controller = InteracterPilotRmBp()

        spectrum_amplitude_figure = grapher.get_figure(properties_freq_amp_figure)
        self._spectrum_amp_frame = framer.get_figure_frame(
            spectrum_amplitude_figure, properties_freq_amp_frame
        )

        spectrum_phase_figure = grapher.get_figure(properties_freq_phase_figure)
        self._spectrum_phase_frame = framer.get_figure_frame(
            spectrum_phase_figure, properties_freq_phase_frame
        )

        self._checkbox_signals_frame = framer.get_checkbox_frame(
            properties_checkbox_signals_frame, controller
        )
        self._checkbox_types_signals_frame = framer.get_checkbox_frame(
            properties_checkbox_type_signals_frame, controller
        )

        main_sweep_figure = grapher.get_figure(properties_main_sweep_figure)
        self._main_sweep_frame = framer.get_figure_frame(
            main_sweep_figure, properties_main_sweep_frame
        )

        self._dropdown_images_spectrogram = framer.get_dropdown_frame(
            properties_dropdown_images_spectrogram_frame, controller
        )

        spectrogram_images_figure = grapher.get_figure(properties_spectrogram_figure)
        self._spectrogram_images_frame = framer.get_figure_frame(
            spectrogram_images_figure, properties_spectrogram_frame
        )

        result_frame = self._spectrum_amp_frame.h_add(self._spectrum_phase_frame)
        checkbox_section = self._checkbox_signals_frame.h_add(
            self._checkbox_types_signals_frame
        )
        result_frame = result_frame.v_add(checkbox_section)
        result_frame = result_frame.v_add(self._main_sweep_frame)
        result_frame = result_frame.v_add(self._dropdown_images_spectrogram)
        result_frame = result_frame.v_add(self._spectrogram_images_frame)

        self._result_frame = result_frame

        controller.set_controled_objects(
            self._dropdown_images_spectrogram,
            self._main_sweep_frame,
            self._spectrum_amp_frame,
            self._spectrum_phase_frame,
            self._spectrogram_images_frame,
            self._checkbox_signals_frame,
            self._checkbox_types_signals_frame,
        )

        self._controller = controller

    def add_amp_spectrum(
        self,
        frequency: np.ndarray,
        amp_spectrum: np.ndarray,
        name: str,
        group: str,
        line_properties: Dict[str, Any],
    ) -> None:
        figure = self._spectrum_amp_frame.get_figure()
        figure.add_line(frequency, amp_spectrum, name, group, line_properties)

    def add_phase_sepctrum(
        self,
        frequency: np.ndarray,
        phase_spectrum: np.ndarray,
        name: str,
        group: str,
        line_properties: Dict[str, Any],
    ) -> None:

        figure = self._spectrum_phase_frame.get_figure()
        figure.add_line(frequency, phase_spectrum, name, group, line_properties)

    def add_choise_sweep(
        self, choise_name: str, group: str, choise_properties: Dict[str, Any]
    ) -> None:

        self._checkbox_signals_frame.add_choise(
            choise_name, group, choise_properties, self._controller
        )

    def add_choise_type_signal(
        self, choise_name: str, group: str, choise_properties: Dict[str, Any]
    ) -> None:
        self._checkbox_types_signals_frame.add_choise(
            choise_name, group, choise_properties, self._controller
        )

    def add_main_sweep_signals(
        self,
        time: np.ndarray,
        amplitude: np.ndarray,
        name: str,
        group: str,
        line_properties: Dict[str, Any],
    ) -> None:

        figure = self._main_sweep_frame.get_figure()
        figure.add_line(time, amplitude, name, group, line_properties)

    def add_choise_dropdown_spectrogram(self, choise_name: str) -> None:
        self._dropdown_images_spectrogram.add_choise(choise_name)

    def add_line_spectrogram(
        self,
        time_spectrogram: np.ndarray,
        frequency_spectrogram: np.ndarray,
        line_name: str,
        group: str,
        line_properties: Dict[str, Any],
    ) -> None:

        figure = self._spectrogram_images_frame.get_figure()
        figure.add_line(
            time_spectrogram, frequency_spectrogram, line_name, group, line_properties
        )

    def add_image_spectrogram(
        self,
        time: np.ndarray,
        frequency: np.ndarray,
        spectrogram: np.ndarray,
        name: str,
        group: str,
        image_properties: Dict[str, Any],
    ) -> None:

        figure = self._spectrogram_images_frame.get_figure()
        figure.add_image(time, frequency, spectrogram, name, group, image_properties)

    def show(self) -> Any:
        return self._result_frame.get_output()



from typing import Dict, Union

from ..base_view.abc_common_framer_grapher.figure import CommonFigure
from ..base_view.abc_common_framer_grapher.frame import (
    CommonCheckBox,
    CommonCheckBoxFrame,
    CommonDropdownFrame,
    CommonFigureFrame,
)

from ..base_view.abc_common_framer_grapher.controller import CommonController


class InteracterPlotViewSweep(CommonController):
    def __init__(self) -> None:
        self._signal_frame = None
        self._checkbox_signals_frame = None
        self._spectrum_amplitude_frame = None
        self._spectrum_phase_frame = None
        self._sweep_frame = None
        self._spectrogram_dropdown_frame = None
        self._spectrogram_images_frame = None
        self._envelop_frame = None
        self._checkbox_sweep_frame = None

    def set_controlled_object(
        self,
        signal_frame: CommonFigureFrame[CommonFigure],
        checkbox_signals_frame: CommonCheckBoxFrame,
        spectrum_amplitude_frame: CommonFigureFrame[CommonFigure],
        spectrum_phase_frame: CommonFigureFrame[CommonFigure],
        sweep_frame: CommonFigureFrame[CommonFigure],
        spectrogram_dropdown_frame: CommonDropdownFrame,
        spectrogram_images_frame: CommonFigureFrame[CommonFigure],
        envelop_frame: CommonFigureFrame[CommonFigure],
        checkbox_sweep_frame: CommonCheckBoxFrame,
    ) -> None:
        self._signal_frame = signal_frame
        self._checkbox_signals_frame = checkbox_signals_frame
        self._spectrum_amplitude_frame = spectrum_amplitude_frame
        self._spectrum_phase_frame = spectrum_phase_frame
        self._sweep_frame = sweep_frame
        self._spectrogram_dropdown_frame = spectrogram_dropdown_frame
        self._spectrogram_images_frame = spectrogram_images_frame
        self._envelop_frame = envelop_frame

        self._checkbox_sweep_frame = checkbox_sweep_frame

    def widget_notify(self, obj: Union[CommonCheckBox, CommonDropdownFrame]):
        def listner(value: Dict):
            if isinstance(obj, CommonCheckBox):
                self._set_line_by_checkbox(obj, value["new"])
                self._spectrogram_dropdown_frame
            if obj is self._spectrogram_dropdown_frame:
                self._set_image_spectrogram(value["new"], True)
                self._set_image_spectrogram(value["old"], False)

        return listner

    def _set_line_by_checkbox(self, check_box: CommonCheckBox, value: bool) -> None:
        fig_frames = [
            self._signal_frame,
            self._spectrum_amplitude_frame,
            self._spectrum_phase_frame,
            self._sweep_frame,
            self._spectrogram_images_frame,
            self._envelop_frame,
        ]
        for fig_frame in fig_frames:
            line = fig_frame.get_figure().get_line(check_box.get_name())
            if line:
                line.set_visible(value)

    def _set_image_spectrogram(self, image_name: str, value: bool) -> None:
        image = self._spectrogram_images_frame.get_figure().get_image(image_name)
        if image:
            image.set_visible(value)



import copy
from ..framemakers.ipywidgetframer.figwidget.bokehfig import BokehFigureMaker
from ..framemakers.ipywidgetframer.ipywidgetframe import WidgetFrameMaker
from .source_view_builder import ViewSource
from ..framemakers.ipywidgetframer.default_ipywidget_properties import (
    get_checkbox_frame,
    get_dropdown_frame,
    get_figure_frame,
)
from ..framemakers.ipywidgetframer.figwidget.default_bokeh_properties import (
    get_figure_window,
)

from ..framemakers.ipywidgetframer.figwidget.bokehfig import default_line_property


def get_view_source_bokeh_ipywidget(
    reaction_mass=1.0, limits=None, k_width=1.0, k_height=1.0
):

    sweep_color = "blue"
    displacement_color = "green"
    envelop_color = "red"

    line_properties_sweep = copy.deepcopy(default_line_property)
    line_properties_sweep["line_color"] = sweep_color

    line_properties_envelope = copy.deepcopy(default_line_property)
    line_properties_envelope["line_color"] = envelop_color

    line_properties_displacement = copy.deepcopy(default_line_property)
    line_properties_displacement["line_color"] = displacement_color

    line_properties_limits = copy.deepcopy(default_line_property)
    line_properties_limits["line_color"] = "yellow"

    figure_window = get_figure_window(
        "Signals", "Time, s", "Force, N", k_width * 0.8, k_height
    )
    figure_window.update(
        {"extra_range_y_name": "displacement", "y_axis_label_2": "Position, m"}
    )
    figure_frame = get_figure_frame(k_width * 0.8, k_height)
    checbox_frame = get_checkbox_frame(
        "Sweep signals:", "No sweep signals!", k_width * 0.2, k_height
    )
    dropdown_frame = get_dropdown_frame("Sweep signals:", "No sweep signals!")
    line_properties_displacement["y_range_name"] = figure_window["extra_range_y_name"]

    return ViewSource(
        BokehFigureMaker,
        WidgetFrameMaker,
        figure_window,
        figure_frame,
        dropdown_frame,
        checbox_frame,
        line_properties_sweep,
        line_properties_envelope,
        line_properties_displacement,
        line_properties_limits,
        reaction_mass,
        limits,
    )



import unittest

import numpy as np

from ..defaults import names as dn
from ..relation_header import HeaderRelation
from ....math_signals.defaults.base_structures import MathOperation


class TestHeaderRelation(unittest.TestCase):
    """Test headers."""

    def check_equal(self, r_named: HeaderRelation, name, category):
        self.assertEqual(str(r_named), name)
        self.assertEqual(r_named.category, category)

    def setUp(self) -> None:
        self.name = "test_name"
        self.category = "test_category"
        self.nr = HeaderRelation(self.name, self.category)
        self.dname = "R_1"
        self.dcategory = "relation"
        self.dnr = HeaderRelation()

    def tearDown(self) -> None:
        HeaderRelation._quantity = 1

    def test_init(self):

        self.check_equal(self.nr, self.name, self.category)

    def test_select_data(self):

        x_start = 0
        x_end = 1

        nr = self.nr.select_data(x_start, x_end)
        self.check_equal(
            nr,
            dn.make_default_name_select_data(self.nr, x_start, x_end),
            self.dcategory,
        )

        nr2 = self.nr.select_data(x_start, x_end, self.name, self.category)
        self.check_equal(nr2, self.name, self.category)

    def test_exp(self):
        nr = self.nr.exp()
        self.check_equal(nr, dn.make_default_name_exp(self.nr), self.nr.category)

        nr2 = self.nr.exp(self.name, self.category)
        self.check_equal(nr2, self.name, self.category)

    def test_diff(self):
        nr = self.nr.diff()
        self.check_equal(
            nr,
            dn.make_default_diff_name(self.nr),
            dn.make_diff_category_name(self.nr.category),
        )

        nr2 = self.nr.exp(self.name, self.category)
        self.check_equal(nr2, self.name, self.category)

    def test_integrate(self):
        nr = self.nr.integrate()
        self.check_equal(
            nr,
            dn.make_default_integrate_name(self.nr),
            dn.make_integrate_category_name(self.nr.category),
        )

        nr2 = self.nr.diff(self.name, self.category)
        self.check_equal(nr2, self.name, self.category)

    def test_interpolate_extrapolate(self):
        new_x = np.array([0.0, 1.0])
        nr = self.nr.interpolate_extrapolate(new_x)
        self.check_equal(
            nr, dn.make_default_interpolate_name(self.nr, new_x), self.nr.category
        )

        nr2 = self.nr.interpolate_extrapolate(new_x, self.name, self.category)
        self.check_equal(nr2, self.name, self.category)

    def test_shift(self):
        x_shift = 10.0

        nr = self.nr.shift(x_shift)
        self.check_equal(
            nr, dn.make_default_shift_name(self.nr, x_shift), self.nr.category
        )

        nr2 = self.nr.shift(x_shift, self.name, self.category)
        self.check_equal(nr2, self.name, self.category)

    def test_operation(self):
        list_operation = [
            MathOperation.ADD,
            MathOperation.RADD,
            MathOperation.SUB,
            MathOperation.RSUB,
            MathOperation.MUL,
            MathOperation.RMUL,
            MathOperation.TRUEDIV,
            MathOperation.RTRUEDIV,
            MathOperation.POW,
            MathOperation.RPOW,
        ]
        for k in list_operation:
            for m in [1.0, self.nr]:
                with self.subTest(k=k, m=m):
                    result = getattr(self.nr, k.value)(m)
                    self.check_equal(
                        result,
                        dn.make_default_names_operations(self.nr, m, k),
                        self.nr.category,
                    )

    def test_equalize(self):
        new_x = np.array([0.0, 1.0])
        nr, nr2 = HeaderRelation.equalize(self.nr, self.nr, new_x)

        self.check_equal(
            nr, dn.make_default_interpolate_name(self.nr, new_x), self.nr.category
        )
        self.check_equal(
            nr2, dn.make_default_interpolate_name(self.nr, new_x), self.nr.category
        )

        nr3, nr4 = HeaderRelation.equalize(
            self.nr, self.nr, new_x, self.name, self.name, self.category, self.category
        )
        self.check_equal(nr3, self.name, self.category)
        self.check_equal(nr4, self.name, self.category)

    def test_correlate(self):
        nr = HeaderRelation.correlate(self.nr, self.nr)
        self.check_equal(
            nr, dn.make_default_name_correlation(self.nr, self.nr), self.nr.category
        )
        nr2 = HeaderRelation.correlate(self.nr, self.nr, self.name, self.category)
        self.check_equal(nr2, self.name, self.category)

    def test_correlate(self):
        nr = HeaderRelation.convolve(self.nr, self.nr)
        self.check_equal(
            nr, dn.make_default_name_convolution(self.nr, self.nr), self.nr.category
        )
        nr2 = HeaderRelation.convolve(self.nr, self.nr, self.name, self.category)
        self.check_equal(nr2, self.name, self.category)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sweep_design',
 'sweep_design.config',
 'sweep_design.dataio',
 'sweep_design.dataio.read',
 'sweep_design.dataio.write',
 'sweep_design.math_signals',
 'sweep_design.math_signals.defaults',
 'sweep_design.math_signals.prepared_sweeps',
 'sweep_design.math_signals.test',
 'sweep_design.math_signals.utility_functions',
 'sweep_design.named_signals',
 'sweep_design.named_signals.defaults',
 'sweep_design.named_signals.header_signals',
 'sweep_design.named_signals.header_signals.defaults',
 'sweep_design.named_signals.header_signals.test',
 'sweep_design.view',
 'sweep_design.view.base_view',
 'sweep_design.view.base_view.abc_common_framer_grapher',
 'sweep_design.view.framemakers',
 'sweep_design.view.framemakers.ipywidgetframer',
 'sweep_design.view.framemakers.ipywidgetframer.defaults',
 'sweep_design.view.framemakers.ipywidgetframer.figwidget',
 'sweep_design.view.view_general',
 'sweep_design.view.view_pilot_rm_bp',
 'sweep_design.view.view_source',
 'sweep_design.view.view_sweep']

package_data = \
{'': ['*']}

install_requires = \
['EMD-signal>=1.2.2,<2.0.0',
 'bokeh>=2.4.2,<3.0.0',
 'ipywidgets>=7.6.5,<8.0.0',
 'jupyter-bokeh>=3.0.4,<4.0.0',
 'jupyterlab>=3.2.9,<4.0.0',
 'loguru>=0.6.0,<0.7.0',
 'matplotlib>=3.5.1,<4.0.0',
 'notebook>=6.4.8,<7.0.0',
 'numpy>=1.22.3,<2.0.0',
 'scipy>=1.8.0,<2.0.0',
 'screeninfo>=0.8,<0.9']

setup_kwargs = {
    'name': 'sweep-design',
    'version': '0.1.0',
    'description': 'The project is designed to create simple and complex sweep signals.',
    'long_description': None,
    'author': 'Vladislav',
    'author_email': 'serebraykov.vs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



import numpy as np
from loguru import logger


def get_time(end=10, dt=0.001, start=0):
    """Create and return time axis."""
    time = np.linspace(start, end, int((end - start) / dt) + 1)
    logger.info(
        "Time axis was created with next parameters: start - {}, end - {} dt - {}",
        time[0],
        time[-1],
        dt,
    )
    return time



from ....base_view.default_screen_properties import get_screen_params

result = get_screen_params()

width, height = result["width"], result["height"]


def get_figure_window(
    title: str, x_axis_label: str, y_axis_label: str, k_width=1.0, k_height=1.0
):
    figure_window = {
        "width": width * k_width * 0.95,
        "height": height * k_height * 0.95,
        "title": title,
        "x_axis_label": x_axis_label,
        "y_axis_label": y_axis_label,
    }
    return figure_window



signal_window = {
    "width": 1000,
    "height": 300,
    "title": "Signal",
    "x_axis_label": "Time, s",
    "y_axis_label": "Amplitude",
}

spectrum_window_amp = {
    "width": 1000,
    "height": 300,
    "title": "Amplitude spectrum",
    "x_axis_label": "Frequency, Hz",
    "y_axis_label": "Amplitude",
}

spectrum_window_phase = {
    "width": 1000,
    "height": 300,
    "title": "Phase spectrum",
    "x_axis_label": "Frequency, Hz",
    "y_axis_label": "Phase",
}

sweep_window = {
    "width": 1000,
    "height": 300,
    "title": "Sweep",
    "x_axis_label": "Time, s",
    "y_axis_label": "Amplitude",
}

spectrogramma_window = {
    "width": 1000,
    "height": 300,
    "title": "Speectrogram",
    "x_axis_label": "Time, s",
    "y_axis_label": "Frequency, Hz",
}

f_t_window = {
    "width": 1000,
    "height": 300,
    "title": "Frequency modulation",
    "x_axis_label": "Time, s",
    "y_axis_label": "Frequency, Hz",
}

a_t_window = {
    "width": 1000,
    "height": 300,
    "title": "Amplitude modulation",
    "x_axis_label": "Time, s",
    "y_axis_label": "Amplitude",
}

default_categories_properties = {
    "amp_spectrum": spectrum_window_amp,
    "phase_spectrum": spectrum_window_phase,
    "signal": signal_window,
    "sweep": sweep_window,
    "spectrogram": spectrogramma_window,
    "f_t": f_t_window,
    "a_t": a_t_window,
}

default_categories_layout = [
    ["signal"],
    ["amp_spectrum"],
    ["phase_spectrum"],
    ["sweep"],
    ["spectrogram"],
    ["f_t"],
    ["a_t"],
]



from typing import Dict, Iterable, Literal, Set, Tuple, Any, Union
import copy
import itertools
import ipywidgets as widgets
import numpy as np

from matplotlib.figure import Figure

from ....base_view.abc_common_framer_grapher.figure import (
    CommonFigure,
    CommonFigureMaker,
    CommonImage,
    CommonLine,
)

import matplotlib.pyplot as plt

import matplotlib.colors as mcolors


PaletteColors = Iterable[Tuple[float, float, float]]

_px = 1 / plt.rcParams["figure.dpi"]
_default_line_properties = {"color": None, "linewidth": 2, "linestyle": "solid"}


_color_palette = itertools.cycle(mcolors.TABLEAU_COLORS)


class MetaFigure(type(widgets.Output), type(CommonFigure)):
    pass


class MatplotlibFigureMaker(CommonFigureMaker):
    def __init__(
        self,
        default_line_properties=_default_line_properties,
        color_palette=_color_palette,
    ) -> None:
        self._default_line_properties = default_line_properties
        self._color_palette = color_palette
        self._default_image_properties = None

    @staticmethod
    def get_figure(figure_properties: Dict[str, Any]) -> "PltFigure":
        fig, axes = plt.subplots(
            constrained_layout=True,
            figsize=(
                0.75 * figure_properties["width"] * _px,
                0.75 * figure_properties["height"] * _px,
            ),
        )

        fig.canvas.header_visible = False
        axes.set_xlabel(figure_properties["x_axis_label"])
        axes.set_ylabel(figure_properties["y_axis_label"])
        axes.set_title(figure_properties["title"])

        return PltFigure(fig, axes)

    def get_next_line_properties(self) -> Dict[str, Any]:
        line_properties = copy.deepcopy(self._default_line_properties)
        line_properties["color"] = next(self._color_palette)
        return line_properties

    def get_next_image_properties(self) -> Dict[str, Any]:
        return "twilight"


class PltFigure(widgets.Output, CommonFigure[Figure], metaclass=MetaFigure):
    def __init__(self, figuer, axes) -> None:
        super().__init__()
        self.figure = figuer
        self.axes = axes
        self.mult_axes = {}
        self._lines = {}
        self._images = {}

        with self:
            plt.show(self.figure)

    def update(self):
        self.figure.canvas.draw()

    def add_extra_axis(self, axis_name: str, axis_label: str, location="right") -> None:
        self.mult_axes[axis_name] = self.axes.twinx()
        self.mult_axes[axis_name].label(axis_label)

    def get_axis(self, axis_name: str = None, type_axis: str = "y"):
        return self.mult_axes[axis_name]

    def add_line(
        self, x, y, unique_name: str, groups: Union[str, Set[str]], line_property: dict
    ) -> "PltLine":
        (line,) = self.axes.plot(x, y, **line_property)
        plt_line = PltLine(line, unique_name, groups, self.update)
        self._lines.update({unique_name: plt_line})
        return plt_line

    def add_image(
        self,
        x: np.ndarray,
        y: np.ndarray,
        s_xy: np.ndarray,
        unique_name: str,
        groups: Set[str],
        image_properties: dict,
    ) -> None:
        extent = (x[0], x[-1], y[0], y[-1])
        image = self.axes.imshow(
            s_xy, extent=extent, origin="lower", aspect="auto", cmap=image_properties
        )
        self._images.update(
            {unique_name: PltImage(image, unique_name, groups, self.update)}
        )

    def add_infty_line(
        self,
        k: np.ndarray,
        axis: Literal["x", "y"],
        unique_name: str,
        groups: Set[str],
        line_properties: Dict[str, Any],
    ) -> None:
        pass

    def show(self):
        plt.show(self.figure)


class PltLine(CommonLine):
    def __init__(self, line, name: str, groups: Union[str, Set[str]], update) -> None:
        super().__init__(line, name, groups)
        self._update = update

    def get_color(self) -> str:
        color = self._line.get_color()
        color = color.split("tab:")[-1]
        return color

    def set_visible(self, is_visible: bool) -> None:

        self._line.set_visible(is_visible)
        self._update()


class PltImage(CommonImage):
    def __init__(self, image, name: str, groups: Set[str], update) -> None:
        super().__init__(image, name, groups)
        self._update = update

    def set_visible(self, is_visible: bool) -> None:

        self._image.set_visible(is_visible)
        self._update()



from typing import Any, Dict
from screeninfo import get_monitors


def get_screen_params(number_screens=1) -> Dict[str, Any]:

    cnt = 1
    for m in get_monitors():
        if cnt == number_screens:
            break
        cnt += 1

    return {"height": m.height, "width": m.width}



from pathlib import Path
from typing import Union

from scipy.io import loadmat

from ...math_signals import Relation
from .defaults import get_read_file_name


def read_mat_file(file_name: Union[Path, str]) -> Relation:
    """Read data from mat file."""
    file_name = get_read_file_name(file_name)
    data = loadmat(file_name)
    print(data)
    columns = [column for column in data.values()]
    return Relation(x=columns[3][0], y=columns[4][0])



from typing import Callable, Union, TypeVar

from .defaults import names as dfn
from .base_header import HeaderBase
from .relation_header import HeaderRelation
from .defaults.methods import make_name, make_category

InName = Union[HeaderBase, str, Callable[[], str]]
NSP = TypeVar("NSP", bound="HeaderSpectrum")


class HeaderSpectrum(HeaderRelation):

    _quantity = 1

    def __init__(self, name: InName = None, category: str = None) -> None:

        if category is None:
            category = "spectrum"

        if name is None:
            name = dfn.make_default_spectrum_name(HeaderSpectrum._quantity)
            HeaderSpectrum._quantity += 1

        super().__init__(name, category)

    def get_signal(self, name: InName = None, category: str = None) -> "HeaderSignal":
        name = make_name(name, dfn.make_default_spectrum2signal_name, self)
        return HeaderSignal(name, category)

    def get_amp_spectrum(
        self, name: InName = None, category: str = None
    ) -> "HeaderSpectrum":
        name = make_name(name, dfn.get_default_spectrum_name, self)
        category = make_category(self, category, dfn.make_default_amp_category_name)
        return HeaderSpectrum(name, category)

    def get_phase_spectrum(
        self, name: InName = None, category: str = None
    ) -> "HeaderSpectrum":
        name = make_name(name, dfn.get_default_spectrum_name, self)
        category = make_category(self, category, dfn.make_default_phase_category_name)
        return HeaderSpectrum(name, category)

    def get_reverse_filter(
        self, name: InName = None, category: str = None
    ) -> "HeaderSpectrum":
        name = make_name(name, dfn.make_default_name_reverse_filter, self)
        category = make_category(self, category)
        return HeaderSpectrum(name, category)

    def add_phase(
        self, other: HeaderBase, name: InName = None, category: str = None
    ) -> "HeaderSpectrum":
        name = make_name(name, dfn.make_default_add_phase_name, self, other)
        category = make_category(self, category)
        return HeaderSpectrum(name, category)

    def sub_phase(
        self, other: HeaderBase, name: InName = None, category: str = None
    ) -> "HeaderSpectrum":
        name = make_name(name, dfn.make_default_sub_phase_name, self, other)
        category = make_category(self, category)
        return HeaderSpectrum(name, category)

    @staticmethod
    def get_spectrum_from_amp_phase(
        amp: HeaderBase, phase: HeaderBase, name: InName = None, category="spectrum"
    ) -> "HeaderSpectrum":
        name = make_name(
            name, dfn.make_default_spectrum_from_amp_phase_name, amp, phase
        )
        return HeaderSpectrum(name, category)


class HeaderSignal(HeaderRelation):

    _quantity = 1
    _make_default_name = dfn.make_default_signal_name

    def __init__(self, name: InName = None, category: str = None) -> None:

        if category is None:
            category = "signal"

        super().__init__(name, category)

    def get_spectrum(
        self, name: InName = None, category="spectrum"
    ) -> "HeaderSpectrum":
        name = make_name(name, dfn.make_default_signal2spectrum_name, self)
        category = make_category(self, category)
        return HeaderSpectrum(name, category)

    def get_amlitude_spectrum(
        self, name: InName = None, category="amp_spectrum"
    ) -> "HeaderSpectrum":
        name = make_name(name, dfn.make_default_spectrum_from_amp_name, self)
        category = make_category(self, category)
        return HeaderSpectrum(name, category)

    def get_phase_spectrum(
        self, name: InName = None, category="phase_spectrum"
    ) -> "HeaderSpectrum":
        name = make_name(name, dfn.make_default_spectrum_from_phase_name, self)
        category = make_category(self, category)
        return HeaderSpectrum(name, category)

    def get_reverse_signal(
        self, name: InName = None, category: str = None
    ) -> "HeaderSignal":
        name = make_name(name, dfn.make_default_name_reverse_signal, self)
        category = make_category(self, category)
        return HeaderSignal(name, category)

    def add_phase(
        self, other: HeaderBase, name: InName = None, category: str = None
    ) -> "HeaderSignal":
        name = make_name(name, dfn.make_default_add_phase_name, self, other)
        category = make_category(self, category)
        return HeaderSignal(name, category)

    def sub_phase(
        self, other: HeaderBase, name: InName = None, category: str = None
    ) -> "HeaderSignal":
        name = make_name(name, dfn.make_default_sub_phase_name, self, other)
        category = make_category(self, category)
        return HeaderSignal(name, category)



from typing import Union, Callable

from .relation_header import HeaderRelation
from .signal_header import HeaderSignal
from .base_header import HeaderBase
from .defaults import names as dfn
from .defaults.methods import make_name, make_category

InName = Union[HeaderBase, str, Callable[[], str]]


class HeaderSweep(HeaderSignal):

    _quantity = 1
    _make_default_name = dfn.default_sweep_name

    def __init__(self, name: InName = None, category: str = None) -> None:

        if category is None:
            category = "sweep"

        super().__init__(name, category)

    def get_a_t(self, name: InName = None, category: str = None) -> HeaderRelation:
        if category is None:
            category = "a_t"
        name = make_name(name, dfn.make_default_a_t_name, self)
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def get_f_t(self, name: InName = None, category: str = None) -> HeaderRelation:
        if category is None:
            category = "f_t"
        name = make_name(name, dfn.make_default_f_t_name, self)
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def get_spectrogram(
        self, name: InName = None, category: str = None
    ) -> HeaderRelation:
        if category is None:
            category = "f_t"
        name = make_name(name, dfn.make_default_spectrogram_name, self)
        category = make_category(self, category)
        return HeaderRelation(name, category)


class HeaderSpectrogram(HeaderBase):

    _quantity = 1

    def __init__(self, name: InName = None, category: str = None) -> None:

        if category is None:
            category = "time_frequebcy"

        if name is None:
            name = dfn.make_spectrogram_name(HeaderSpectrogram._quantity)
            HeaderSpectrogram._quantity += 1

        super().__init__(name, category)

    @property
    def category(self) -> str:
        return self._category

    @category.setter
    def category(self, value) -> None:
        self._category = value

    @property
    def name(self) -> Callable[..., str]:
        return self._name

    @name.setter
    def name(self, value: InName) -> None:
        if callable(value) and isinstance(value(), str):
            self._name = value
        elif isinstance(value, HeaderBase):
            self._name = value._name
        else:
            str(value)

            def call() -> str:
                return str(value)

            self._name = call

    def __str__(self) -> str:
        return self.name()

    def __repr__(self) -> str:
        return f"Name: {self.name()} category: {self.category}"



"""Utility functions."""

from .time_axis import get_time as get_time
from .emd_analyse import get_IMFs_ceemdan as get_IMFs_ceemdan
from .tukey import tukey_a_t as tukey_a_t
from .linear_f_t import f_t_linear_array as f_t_linear_array
from .linear_f_t import f_t_linear_function as f_t_linear_function
from .ftat_functions import simple_freq2time as simple_freq2time
from .ftat_functions import dwell as dwell
from .sweep_correction import (
    correct_sweep_without_window as correct_sweep_without_window,
)
from .sweep_correction import correct_sweep as correct_sweep
from .sweep_correction_source import (
    get_correction_for_source as get_correction_for_source,
)



from typing import Callable, Optional, Union
import numpy as np

from ....math_signals.defaults.base_structures import MathOperation
from ..base_header import HeaderBase

Num = Union[float, int, complex]
InName = Union[HeaderBase, str, Callable[[], str]]


def make_default_names_operations(
    obj1: HeaderBase, obj2: Union[HeaderBase, Num], operation: MathOperation
) -> str:
    default_names_operations = {
        "__add__": "({0} + {1})",
        "__sub__": "({0} - {1})",
        "__mul__": "({0} * {1})",
        "__truediv__": "({0} / {1})",
        "__pow__": "({0} ** {1})",
        "__radd__": "({0} + {1})",
        "__rsub__": "({0} - {1})",
        "__rmul__": "({0} * {1})",
        "__rtruediv__": "({0} / {1})",
        "__rpow__": "({0}) ** {1})",
    }
    return default_names_operations[operation.value].format(obj1, obj2)


def make_default_name_select_data(obj: HeaderBase, x_start: Num, x_end: Num) -> str:
    return "{0} [{1}-{2}]".format(obj.name(), x_start, x_end)


def make_default_name_exp(obj: HeaderBase) -> str:
    return "exp({0})".format(obj.name())


def make_default_diff_name(obj: HeaderBase) -> str:
    return "diff({0})".format(obj.name())


def make_default_integrate_name(obj: HeaderBase) -> str:
    return "integrate({0})".format(obj.name())


def make_default_shift_name(obj: HeaderBase, x_shift: Num) -> str:
    return "Shift({0}) by {1}".format(obj.name(), x_shift)


def make_default_interpolate_name(obj: HeaderBase, new_x: np.ndarray) -> str:
    return "inerp({0} {1}-{2} dx:{3})".format(
        obj.name(), new_x[0], new_x[-1], new_x[1] - new_x[0]
    )


def make_default_name_correlation(obj1: HeaderBase, obj2: HeaderBase) -> str:
    return "(Corr({0} & {1}))".format(obj1._name(), obj2._name())


def make_default_name_convolution(obj1: HeaderBase, obj2: HeaderBase) -> str:
    return "(Conv({0} & {1}))".format(obj1, obj2)


def make_default_relation_name(name: int) -> str:
    return "R_{0}".format(name)


def make_diff_category_name(name: str) -> str:
    return "diff_{0}".format(name)


def make_integrate_category_name(name: str) -> str:
    return "integrate_{0}".format(name)


# ==============================================================================


def get_default_spectrum_name(obj: HeaderBase) -> str:
    return "{0}".format(obj)


def make_default_spectrum2signal_name(obj: HeaderBase) -> str:
    return "({0})toS".format(obj)


def make_default_signal2spectrum_name(obj: HeaderBase) -> str:
    return "({0})toSP".format(obj)


def make_default_spectrum_from_amp_phase_name(
    obj_amp: HeaderBase, obj_phase: HeaderBase
) -> str:
    return "SP(amp({0}) phase({1}))".format(obj_amp, obj_phase)


def make_default_spectrum_from_phase_name(obj_phase: HeaderBase) -> str:
    return "SP(phase({0}))".format(obj_phase)


def make_default_spectrum_from_amp_name(obj_amp: HeaderBase) -> str:
    return "SP(amp({0}))".format(obj_amp)


def make_default_name_reverse_filter(obj: HeaderBase) -> str:
    return "reverse filter ({0})".format(obj)


def make_default_name_reverse_signal(obj: HeaderBase) -> str:
    return "recerse signal ({0})".format(obj)


def make_default_add_phase_name(obj1: HeaderBase, obj2: HeaderBase) -> str:
    return "{0}+phase({1})".format(obj1, obj2)


def make_default_sub_phase_name(obj1: HeaderBase, obj2: HeaderBase) -> str:
    return "{0}-phase({1})".format(obj1, obj2)


def make_default_spectrum_name(name) -> str:
    return "SP_{0}".format(name)


def make_default_signal_name(name) -> str:
    return "S_{0}".format(name)


def make_default_phase_category_name(name) -> str:
    return "phase_{0}".format(name)


def make_default_amp_category_name(name) -> str:
    return "amp_{0}".format(name)


# #============================================================================


def default_sweep_name(name) -> str:
    return "SW_{0}".format(name)


def make_default_a_t_name(obj: HeaderBase) -> str:
    return "a_t({0})".format(obj)


def make_default_f_t_name(obj: HeaderBase) -> str:
    return "f_t({0})".format(obj)


def make_default_spectrogram_name(obj: HeaderBase) -> str:
    return "sp_f_t({0})".format(obj)


# ==============================================================================


def make_spectrogram_name(name) -> str:
    return "Spectrogram_{0}".format(name)


# ==============================================================================


def default_name_uncalc_sweep(name) -> str:
    return "USW_{0}".format(name)


def default_name_aprior_uncalc_sweep(name) -> str:
    return "AUSW_{0}".format(name)


def default_name_call_uncalc_sweep(dt: float, T: float, obj: HeaderBase) -> str:
    return "{0} dt: {1} T: {2}".format(obj, dt, T)



"""The configuration of project.

> The configuration is defined by three classes: `Config`
(`sweep_design.config.config.Config`),
`SweepConfig` (`sweep_design.config.sweep_config.SweepConfig`) and `NamedConfig` 
(`sweep_design.config.named_config.NamedConfig`).

> Using three config files to avoid the circular import error.

> Detailed description of the attributes in each of the files.
"""

from .config import Config as Config



from typing import Optional, TypeVar, Union, Type

import numpy as np
from numpy.typing import NDArray

from ..config import Config
from .math_relation import Relation
from .defaults.base_structures import ConvertingError


Num = Union[float, int, complex]
R = TypeVar("R", bound=Relation)
SP = TypeVar("SP", bound="Spectrum")
S = TypeVar("S", bound="Signal")
SPRN = Union["Spectrum", "Relation", Num]
SSPR = Union["Spectrum", "Signal", "Relation"]
SSPRN = Union["Spectrum", "Signal", "Relation", Num]


def _input2spectrum_operation(inp: SSPRN) -> Union["Relation", "Spectrum", Num]:
    if isinstance(inp, Signal):
        return inp.get_spectrum()
    elif isinstance(inp, (Spectrum, Relation, int, float, complex)):
        return inp
    else:
        raise ConvertingError(type(inp), Spectrum)


def _input2spectrum(inp: SSPR) -> "Spectrum":
    if isinstance(inp, Signal):
        return inp.get_spectrum()

    elif isinstance(inp, Spectrum):
        return inp

    elif isinstance(inp, Relation):
        return Spectrum(inp)
    else:
        raise ConvertingError(type(inp), Spectrum)


class Spectrum(Relation):
    """A class that describes the Fourier spectrum of a signal.

    The `Spectrun` class derived from the `Relation` class.

    **Properties**:

    > **frequency**: `Union[Relation, NDArray]`
    An instance of Relation class or inherited from it, or array_like object
    containing numbers (real or complex).

    > **spectrum_amplitude**: `Oprional[NDArray]`
    None or array_like object containing numbers (real or complex).

    > **df**: `float` = `None`
    Sample rate of frequency-axis

    To convert the spectrum into a signal, the method defined in the Config
    class is used. (Config.spectrum2signal_method). Current method can be
    overridden by own in `Config` class.

    When performing arithmetic operations on instances of the `Signal` class,
    an instance of the `Spectrum` class will be extracted from
    the `Signal` instance, and arithmetic operations will be performed
    on this instance. An instance of `Relation` class will be converted into
    the instance of `Spectrum` class.
    """

    def __init__(
        self,
        frequency: Union[Relation, NDArray],
        spectrum_amplitude: NDArray = None,
        df: float = None,
        **kwargs
    ) -> None:
        super().__init__(frequency, spectrum_amplitude, df)
        self._spectrum2signal_method_default = Config.spectrum2signal_method
        self.signal: Optional[Signal] = None

    @property
    def frequency(self):
        return self.x

    @property
    def spectrum_amplitude(self):
        return self.y

    @property
    def df(self):
        return self.dx

    @df.setter
    def df(self, value: float):
        self.dx = value

    def get_signal(self, recalculate=False, start_time: float = None) -> "Signal":
        """Compute the signal from the spectrum."""

        if self.signal is None or recalculate:
            time, amplitude = self._spectrum2signal_method_default(
                self._x, self._y, start_time
            )
            self.signal = Signal(time, amplitude)
        return self.signal

    def get_amp_spectrum(self: R, **kwargs) -> "Relation":
        """Amplitude spectrum.

        Calculate the relationship between the frequency and the absolute
        value of the spectrum amplitude."""

        x, y = self.get_data()
        return Relation(x, np.abs(y))

    def get_phase_spectrum(self: R, **kwargs) -> "Relation":
        """Calculate the relationship between frequency and phase of the spectrum."""
        x, y = self.get_data()
        return Relation(x, np.unwrap(np.angle(y)))

    def get_reverse_filter(
        self: SP,
        percent: Union[float, int] = 5.0,
        subtrack_phase=True,
        frequency_start: float = None,
        frequency_end: float = None,
        **kwargs
    ) -> SP:
        """Calculate filter of reversed signal.

        **Properties**:
        > **percent**: `Union[float, int]`
        level of added white noise in percent

        **subtrack_phase**: True
        If True performs phase subtraction,
        If False succeeds, add the phase.

        **frequency_start**: `float`**
        Start frequency.

        **frequency_end**: `float`
        End frequency.

        """
        spectrum = self.select_data(frequency_start, frequency_end)
        abs_spectrum = spectrum.get_amp_spectrum()
        abs_spectrum = abs_spectrum + abs_spectrum.max() * percent / 100
        reversed_abs_spectrum = 1 / abs_spectrum

        if subtrack_phase:
            phase_spectrum = -1 * spectrum.get_phase_spectrum()
        else:
            phase_spectrum = 1 * spectrum.get_phase_spectrum()

        result_spectrum = type(self).get_spectrum_from_amp_phase(
            reversed_abs_spectrum, phase_spectrum, **kwargs
        )
        return result_spectrum

    def add_phase(self: SP, other: SSPR, **kwargs) -> SP:
        sp_other = _input2spectrum(other)
        return type(self).get_spectrum_from_amp_phase(
            self.get_amp_spectrum(),
            self.get_phase_spectrum() + sp_other.get_phase_spectrum(),
            **kwargs
        )

    def sub_phase(self: SP, other: SSPR, **kwargs) -> SP:
        sp_other = _input2spectrum(other)
        return type(self).get_spectrum_from_amp_phase(
            self.get_amp_spectrum(),
            self.get_phase_spectrum() - sp_other.get_phase_spectrum(),
            **kwargs
        )

    @classmethod
    def get_spectrum_from_amp_phase(
        cls: Type[SP], s1: Relation, s2: Relation, **kwargs
    ) -> SP:
        """Calculate of the spectrum from the amplitude and frequency spectrum.

        The spectrum is calculated through the amplitude and phase spectrum
        using the formula abs*exp(1j*phase)."""

        return cls(s1 * ((1.0j * s2).exp()), **kwargs)

    @classmethod
    def convolve(cls: Type[SP], r1: SSPR, r2: SSPR, **kwargs) -> SP:
        sp_r1 = _input2spectrum(r1)
        sp_r2 = _input2spectrum(r2)
        return super().convolve(sp_r1, sp_r2, **kwargs)

    @classmethod
    def correlate(cls: Type[SP], r1: SSPR, r2: SSPR, **kwargs) -> SP:
        sp_r1 = _input2spectrum(r1)
        sp_r2 = _input2spectrum(r2)
        return super().correlate(sp_r1, sp_r2, **kwargs)

    def __add__(self: SP, a: SSPRN, **kwargs) -> SP:
        r_a = _input2spectrum_operation(a)
        return super().__add__(r_a, **kwargs)

    def __sub__(self: SP, a: SSPRN, **kwargs) -> SP:
        r_a = _input2spectrum_operation(a)
        return super().__sub__(r_a, **kwargs)

    def __mul__(self: SP, a: SSPRN, **kwargs) -> SP:
        r_a = _input2spectrum_operation(a)
        return super().__mul__(r_a, **kwargs)

    def __truediv__(self: SP, a: SSPRN, **kwargs) -> SP:
        r_a = _input2spectrum_operation(a)
        return super().__truediv__(r_a, **kwargs)

    def __pow__(self: SP, a: SSPRN, **kwargs) -> SP:
        r_a = _input2spectrum_operation(a)
        return super().__pow__(r_a, **kwargs)


def _inp2signal_operation(inp: SSPRN) -> Union["Relation", "Signal", Num]:
    if isinstance(inp, Spectrum):
        return inp.get_signal()
    elif isinstance(inp, (Signal, Relation, int, complex, float)):
        return inp
    else:
        raise ConvertingError(type(inp), Signal)


def _inp2signal(inp: SSPR) -> "Signal":
    if isinstance(inp, Spectrum):
        return inp.get_signal()

    elif isinstance(inp, Signal):
        return inp

    elif isinstance(inp, Relation):
        return Signal(inp)
    else:
        raise ConvertingError(type(inp), Signal)


class Signal(Relation):
    """Class describing some kind of signal.

    The `Signal` class inherits the `Relation` class.

    **Properties**:
    > **time**: `Union[Relation, NDArray]`
    An instance of Relation class or inherited from it, or array_like object
    containing numbers (real or complex).

    > **amplitude**: `NDArray`
    None or array_like object containing numbers (real or complex).

    > **dt**: `float` = `None`
    Sample rate of time-axis

    To convert the signal into a spectrum, the method defined in the `Config`
    class is used. (Config.signal2spectrum_method). Current method can be
    overridden by own in Config class.

    When performing arithmetic operations on instances of the Spectrum class,
    an instance of the `Singal` class will be extracted from
    the `Spectrum` instance, and arithmetic operations will be performed
    on this instance. An instance of `Relation` class will be converted into
    the instance of `Signal` class.

    """

    def __init__(
        self,
        time: Union[Relation, np.ndarray],
        amplitude: np.ndarray = None,
        dt: float = None,
    ) -> None:

        self._signal2spectrum_method_default = Config.signal2spectrum_method
        super().__init__(time, amplitude, dt)
        self._spectrum: Optional[Spectrum] = None

    def get_spectrum(self, recalculate=False, is_start_zero=False) -> "Spectrum":

        if self._spectrum is None or recalculate:
            f, a = self._signal2spectrum_method_default(*self.get_data(), is_start_zero)
            self._spectrum = Spectrum(f, a)

        return self._spectrum

    def get_amplitude_spectrum(
        self, recalculate=False, is_start_zero=False
    ) -> Relation:
        return self.get_spectrum(recalculate, is_start_zero).get_amp_spectrum()

    def get_phase_spectrum(self, recalculate=False, is_start_zero=False) -> Relation:
        return self.get_spectrum(recalculate, is_start_zero).get_phase_spectrum()

    def get_reverse_signal(
        self: S,
        percent: Union[float, int] = 5.0,
        subtrack_phase: bool = True,
        frequency_start: float = None,
        frequency_end: float = None,
        **kwargs
    ) -> S:
        """Calculate reversed signal.

        **Properties**:

        > **percent**: `Union[float, int]`
        level of added white noise in percent

        > **subtrack_phase**: True
        If True performs phase subtraction,
        If False succeeds, add the phase.

        > **frequency_start**: float.
        Start frequency.

        > **frequency_end**: float
        End frequency.

        """
        signal = (
            self.get_spectrum()
            .get_reverse_filter(percent, subtrack_phase, frequency_start, frequency_end)
            .get_signal()
        )

        return type(self)(signal, **kwargs)

    def add_phase(self: S, other: SSPR, **kwargs) -> S:
        sp_other = _input2spectrum(other)
        self_spectrum = self.get_spectrum()
        new_spectrum = Spectrum.get_spectrum_from_amp_phase(
            self_spectrum.get_amp_spectrum(),
            self_spectrum.get_phase_spectrum() + sp_other.get_phase_spectrum(),
        )
        return type(self)(new_spectrum.get_signal(), **kwargs)

    def sub_phase(self: S, other: SSPR, **kwargs) -> S:
        sp_other = _input2spectrum(other)
        self_spectrum = self.get_spectrum()
        new_spectrum = Spectrum.get_spectrum_from_amp_phase(
            self_spectrum.get_amp_spectrum(),
            self_spectrum.get_phase_spectrum() - sp_other.get_phase_spectrum(),
        )
        return type(self)(new_spectrum.get_signal(), **kwargs)

    @classmethod
    def convolve(cls: Type[S], r1: SSPR, r2: SSPR, **kwargs) -> S:
        s_r1 = _inp2signal(r1)
        s_r2 = _inp2signal(r2)
        return cls(super().convolve(s_r1, s_r2), **kwargs)

    @classmethod
    def correlate(cls: Type[S], r1: SSPR, r2: SSPR, **kwargs) -> S:
        s_r1 = _inp2signal(r1)
        s_r2 = _inp2signal(r2)
        return cls(super().correlate(s_r1, s_r2), **kwargs)

    def __add__(self: S, a: SSPRN, **kwargs) -> S:
        s_a = _inp2signal_operation(a)
        return super().__add__(s_a, **kwargs)

    def __sub__(self: S, a: SSPRN, **kwargs) -> S:
        s_a = _inp2signal_operation(a)
        return super().__sub__(s_a, **kwargs)

    def __mul__(self: S, a: SSPRN, **kwargs) -> S:
        s_a = _inp2signal_operation(a)
        return super().__mul__(s_a, **kwargs)

    def __truediv__(self: S, a: SSPRN, **kwargs) -> S:
        s_a = _inp2signal_operation(a)
        return super().__truediv__(s_a, **kwargs)

    def __pow__(self: S, a: SSPRN, **kwargs) -> S:
        s_a = _inp2signal_operation(a)
        return super().__pow__(s_a, **kwargs)
